--     __           _         _            _   _
--  /\ \ \__ _ _ __| | ____ _| | ___ _ __ | |_(_) ___
-- /  \/ / _` | '__| |/ / _` | |/ _ \ '_ \| __| |/ __|
--/ /\  / (_| | |  |   < (_| | |  __/ |_) | |_| | (__
--\_\ \/ \__,_|_|  |_|\_\__,_|_|\___| .__/ \__|_|\___|
--                                  |_|
-- /Narkaleptic/              narkaleptic@mainelan.net

function ZombieDefense1:StartActivity()

	print( "Zombie Defense initializing!" );

	self.FightStage = { NOFIGHT = 0, FIGHT = 1, AFTERMATH = 2, FIGHTSTAGECOUNT = 3 };
	self.Stage = self.FightStage.NOFIGHT;

	-- Timer to only run the updates every 250ms
	self.UpdateTimer = Timer();

	-- A timer to prevent overriding important messages
	self.MessageTimer = Timer();
	self.ImportantMessageTimer = Timer();

	self.GameOver = false;
	self.ZombieSpawnTime = 2500;

	-- Zombie Spawns
	self.ZombieSpawner = Timer();
	self.ZombieSpawnLeft = SceneMan.Scene:GetArea( "LeftSpawn" );
	self.ZombieSpawnRight = SceneMan.Scene:GetArea( "RightSpawn" );

	-- Zombie Stuck Locations
	self.ZombieStuck = SceneMan.Scene:GetArea( "Stuck" );

	-- Zombie Goal
	self.GoalArea = SceneMan.Scene:GetArea( "Goal" );

	-- Player Start
	self.ReinforceArea = SceneMan.Scene:GetArea( "Reinforce" );

	-- Rescue Shuttle
	self.RescueArea = SceneMan.Scene:GetArea( "Rescue" );
	self.RescuePoint = SceneMan:MovePointToGround( self.RescueArea:GetRandomPoint(), 1, 0 );

	-- Prevent purchasing
	ActivityMan:GetActivity():SetTeamFunds( 0, 0 );

	-- List of earned weapons
	self.WeaponLevel = 0;
	self.Weapons = { "Glock" };

	-- Glock -> Desert Eagle -> Pumpgun -> M1 Garand -> Kar98 
	-- Uzi -> AK-47 -> M16 -> M60
	-- Thumper -> RPG-7 -> RPC M17
	
	self.WeaponRewards =
	{
	--  Kills         Announce				        Inventory, ...
		[0] = { New = "Glock", 				Inv = { "Glock" } },
		[8] = { New = "Desert Eagle", 		Inv = { "Desert Eagle" } },
		[12] = { New = "Uzi", 				Inv = { "Desert Eagle", "Uzi" } },
		[20] = { New = "Pumpgun", 			Inv = { "Pumpgun", "Uzi" } },
		[45] = { New = "AK-47", 			Inv = { "Pumpgun", "AK-47" } },
		[65] = { New = "M1 Garand", 		Inv = { "M1 Garand", "AK-47" } },
		[85] = { New = "Thumper", 			Inv = { "M1 Garand", "AK-47", "Thumper" } },
		[100] = { New = "M16", 				Inv = { "M1 Garand", "M16", "Thumper" } },
		[125] = { New = "Kar98", 			Inv = { "Kar98", "M16", "Thumper" } },
		[150] = { New = "M60", 				Inv = { "Kar98", "M60", "Thumper" } },
		[180] = { New = "RPG-7", 			Inv = { "Kar98", "M60", "RPG-7" } },
		[240] = { New = "RPC M17", 			Inv = { "Kar98", "M60", "RPC M17" } }
	};

	self.StoryElements =
	{
		[0] = "An evacuation was called days ago, but nobody has made it through here.  370 were due to depart.",
		[1] = "There appears to be a slight infestation of zombies.\nSome weapons have been ordered, but they'll be slow coming.",
		[2] = "I need to defend this base long enough to figure\nout what is going on.",
		[3] = "Now with some reinforcements, hopefully progress\ncan be made.",
		[4] = "We need to keep our eyes open for clues.",
		[5] = "Nothing yet, it seems.",
		[6] = "That one was a scientist! Dr. Gregore, from his badge.",
		[7] = "Dr. Gregore was running experiments on Workstation Beta\nat the Research Outpost to the East.",
		[8] = "He stored his data on a Blue Chip in the Armory\nat the Research Outpost to the East.",
		[9] = "His partner was due for departure from Outpost 2,\nbut he never arrived.",
		[10] = "Reports from Outpost 2 to the South indicate\nthat the infection is wide spread."
	};

	-- Create player
	local PlayerActor = self:DropDefender();
	PlayerActor.Pos = SceneMan:MovePointToGround( self.ReinforceArea:GetRandomPoint(), 24, 0 )
	self:SetLandingZone( self.RescuePoint, 0 );
	self:SetObservationTarget( self.RescuePoint, Activity.PLAYER_1 );
	self.PlayerActor = PlayerActor;

	-- Configure zombies
	self.ZombieTypes = { "Spewer Thin", "Spewer Medium", "Spewer Fat" };
	self.ZombieWeapons = { "Vomit Gun" };
	self.ZombieBombs = { "Blue Bomb" };

	-- Track active zombie count and kills
	self.ZombieCount = 0;
	self.ZombieSpawned = 0;
	self.ZombieKilled = 0;
	self.ZombieCap = 12;

	-- Track wave status
	self.Wave = 0;
	self.WaveCap = 5;
	self.WaveCapCap = 100;
	self.WaveSpawned = 0;
	self.WaveKilled = 0;
	self.WaveTimer = Timer();
	self.WaveRest = 5000;
	self.WaveEndRest = 5000;

	print( "Zombie Defense has begun!" );

	function _G.Rescue()
		self:CreateRescueCraft();
	end

	function _G.Reinforce()
		self:DropDefender();
	end	
	
	self.WaveTimer:Reset();
	FrameMan:SetScreenText( "\"" .. self.StoryElements[0] .. "\"", Activity.PLAYER_1, 0, -1, false );
end

function ZombieDefense1:PauseActivity()

end


function ZombieDefense1:EndActivity()

	self.GameOver = true;

	if self.WinnerTeam == Activity.TEAM_2 then
 		self:UpdateMessage( self.EndText );
	else
		self:UpdateMessage( "You've survived and defended the base from infection." );
 	end	

	_G.Rescue = nil;
	_G.Reinforce = nil;

end

function ZombieDefense1:UpdateMessage( Message )

	if Message then
		self.Message = Message;
		self.ImportantMessageTimer:Reset();
	elseif self.ImportantMessageTimer:IsPastRealMS( 6000 ) then
		self.Message = "";
	end

	if self.Wave and self.Wave > 0 then
		Message =
			"Wave " .. self.Wave .. " (" .. math.floor( (self.WaveKilled or 0) / (self.WaveCap or 1) * 100 ) .. "%)\n" ..
			"Kills: " .. tostring( self.ZombieKilled or 0 ) .. "\n" ..
			"Weapons: " .. table.concat( self.Weapons, ", " ) .. "\n";
	else
		Message = "";
	end

    FrameMan:ClearScreenText( Activity.PLAYER_1 );
	FrameMan:SetScreenText( Message .. "\n" .. self.Message, Activity.PLAYER_1, 0, -1, false );
	self:ResetMessageTimer( Activity.PLAYER_1 );
	
end

function ZombieDefense1:StartWave()

	print( "Initializing new wave..." );

	-- Prepare the variables for a new wave
	self.WaveCap = math.min( self.WaveCap + 5, self.WaveCapCap );
	self.WaveSpawned = 0;
	self.WaveKilled = 0;
	self.Wave = self.Wave + 1;

	self.Stage = self.FightStage.FIGHT;

	local Wave = self.Wave;

	self.ZombieTypes = { "Spewer Thin" }; -- { "Spewer Medium", "Spewer Fat" };
	self.ZombieWeapons = {}; -- { "Vomit Gun" };
	self.ZombieBombs = {}; -- { "Blue Bomb" };
	self.ZombieRight = false;
	self.ZombieSpawnTime = 2500;
	self.ZombieCap = 6;

	if Wave >= 2 then
		table.insert( self.ZombieTypes, "Spewer Medium" );
		self.ZombieCap = self.ZombieCap + 4;
	end

	if Wave >= 3 then
		table.insert( self.ZombieWeapons, "Vomit Gun" );
		self.ZombieCap = self.ZombieCap + 2;
	end

	if Wave == 3 then
		self.ZombieRight = 0.10;
	elseif Wave > 3 then
		self.ZombieRight = 0.55;
	end

	if Wave >= 4 then
		-- Send a reinforcement if a defender is lost
		local PlayerCount = 0;
		for Defender in MovableMan.Actors do
			if Defender.Team == 0
			and Defender.ClassName == "AHuman"
			and not Defender:IsDead() then
				PlayerCount = PlayerCount + 1;
			end
		end
		if PlayerCount == 1 then
			self:DropDefender();
		end
	end

	if Wave >= 5 then
		self.ZombieSpawnTime = 1500;
		self.ZombieCap = self.ZombieCap + 6;
		table.insert( self.ZombieBombs, "Blue Bomb" );
	end

	if Wave >= 6 then
		self.ZombieCap = self.ZombieCap + 6;
	end

	if Wave >= 7 then
		self.ZombieSpawnTime = 750;
	end

	if Wave >= 8 then
		table.insert( self.ZombieBombs, "Cluster Grenade" );
		table.insert( self.ZombieTypes, "Spewer Fat" );
		self.ZombieCap = self.ZombieCap + 6;
	end

	if Wave >= 9 then
		self.ZombieCap = self.ZombieCap + 4;
	end

	print( "Wave " .. Wave .. " started!" );	
	
	-- Announce new wave
	self:UpdateMessage( "\"" .. self.StoryElements[Wave] .. "\"\n\nWave " .. self.Wave .. " has begun!" );
end

function ZombieDefense1:StopWave()

	print( "Stopping wave..." );

	-- Start countdown until wave end
	self.WaveTimer:Reset();

	-- Each time a zombie is killed, give more time, but don't do anything else.
	if self.Stage ~= self.FightStage.FIGHT then
		return
	end

	self.Stage = self.FightStage.AFTERMATH;

	-- Player has a moment to kill the last few zombies

end

function ZombieDefense1:EndWave()

	print( "Wave ended!" );

	-- Kill off the remaining zombies
	for Zombie in MovableMan.Actors do
		if Zombie.Team == 1 and Zombie.ClassName == "AHuman" then
			Zombie:GibThis();
		end
	end
	self.ZombieCount = 0;

	-- Reset the defender inventories
	for Defender in MovableMan.Actors do
		if Defender.Team == 0 and Defender.ClassName == "AHuman" then
			Defender:DropAllInventory();
		end
	end

	-- Announce end of wave
	self:UpdateMessage( "Wave " .. self.Wave .. " has ended!" );		
	
	-- Start countdown until next wave
	self.WaveTimer:Reset();

	self.Stage = self.FightStage.NOFIGHT;

	-- Call in reinforcements when wave 3 is over
	if self.Wave == 3 then
		self:DropDefender();
	end

	if self.Wave == 9 then
		self:CreateRescueCraft();
	end

end

function ZombieDefense1:CraftEnteredOrbit()

    if self.OrbitedCraft:HasObject( "Gordon" ) then

		self.WinnerTeam = Activity.TEAM_1;
		ActivityMan:EndActivity();

    end

end

function ZombieDefense1:UpdateActivity()

	-- Only think four times per second
	if not self.UpdateTimer:IsPastSimMS( 250 ) then
		return
	end
	self.UpdateTimer:Reset();

	-- Update the HUD once per second
	if self.MessageTimer:IsPastRealMS( 1000 ) then
		self:UpdateMessage();
		self.MessageTimer:Reset();
	end

	-- Remove lingering items
	for Item in MovableMan.Items do
		if Item.ClassName == "HDFirearm" then
			if Item.PresetName == "Vomit Gun" then
				Item.ToDelete = true;
			elseif Item.PresetName == "Blue Bomb" then
				Item.ToSettle = true;
			else
				Item.ToDelete = true;
			end
		end
	end

	-- Do not perform game logic if the game is over
	if self.GameOver then
		return
	end

	-- Count player actors
	local PlayerCount = 0;
	for Defender in MovableMan.Actors do
		if Defender.Team == 0 and not Defender:IsDead() then
			if Defender.ClassName == "AHuman" then
				PlayerCount = PlayerCount + 1;
				self.PlayerActor = Defender
				if Defender:IsInventoryEmpty() then
					self:EquipDefender( Defender );
				end
			elseif Defender.ClassName == "ACRocket" or Defender.ClassName == "ACDropShip"  then
				PlayerCount = PlayerCount + 1;

				-- Make sure any rockets ease down slowly
				self.PlayerActor = Defender
			end
		end
	end

	-- Check for living players
	-- Don't start the game until there is a player
	if PlayerCount == 0 then
		if self.Started then
			self.GameOver = true;
			self.WinnerTeam = Activity.TEAM_2;
			self.EndText = "You have perished.";
			ActivityMan:EndActivity();
		else
			return
		end
	elseif not self.Started then
		self.Started = true
	end

	if PlayerCount == 1 and self.LastPlayerCount ~= 1 then
		--print( "Changing active to " .. tostring( self.PlayerActor ) );
		--self:SwitchToActor( self.PlayerActor, 0, self:GetTeamOfPlayer( 0 ) );
		--self:SetObservationTarget( self.RescuePoint, Activity.PLAYER_1 );

		-- If the last player actor is a rocket or dropship, send it home
		if self.PlayerActor.ClassName == "ACRocket"
		or self.PlayerActor.ClassName == "ACDropShip"  then
			 self.PlayerActor.AIMode = Actor.AIMODE_RETURN;
		end
	end
	self.LastPlayerCount = PlayerCount;

	-- Zombie spawn
	if self.Stage ~= self.FightStage.NOFIGHT
	and self.ZombieSpawner:IsPastSimMS( self.ZombieSpawnTime ) then
		self:SpawnZombie();
	end

	-- Progress waves
	if self.Stage == self.FightStage.NOFIGHT then

		-- After a brief rest, start the wave
		if self.WaveTimer:IsPastRealMS( self.WaveRest ) then
			self:StartWave();
		end

		return;

	elseif self.Stage == self.FightStage.AFTERMATH then

		-- After a moment, or if all zombies are dead, end the wave
		if ( self.WaveKilled >= self.WaveCap )
		or ( self.WaveTimer:IsPastRealMS( self.WaveRest ) ) then
			self:EndWave();
		end

	end

	-- Check for zombies entering the base or stuck
	for Zombie in MovableMan.Actors do
        if Zombie.Team == 1
		and Zombie.ClassName == "AHuman" then
			if SceneMan.Scene:WithinArea( "Goal", Zombie.Pos ) then
				self.GameOver = true;
				self.WinnerTeam = Activity.TEAM_2;
				self.EndText = "The zombies have entered the base.  No rescue team will risk infection.";
				ActivityMan:EndActivity();
			elseif self.ZombieStuck and SceneMan.Scene:WithinArea( "Stuck", Zombie.Pos ) then
				Zombie.Pos = self:PickZombieStart();
			end
		end
    end

end

function ZombieDefense1:CreateRescueCraft()

	print( "Creating rescue craft!" );
	local Craft = CreateACRocket( "Rocket MK2" );
	Craft.Pos = Vector( self.RescueArea:GetRandomPoint().X, -50 )
	Craft.Team = 0;
	Craft.AIMode = Actor.AIMODE_STAY;
	Craft:AddAISceneWaypoint( self.RescueArea:GetRandomPoint() );
	MovableMan:AddActor( Craft );

	-- Remove the bay doors
	for Door in MovableMan.Actors do
		if ( Door.ClassName == "ADoor" )
		and ( Door.PresetName == "Door Rotate Long" ) then
			print( "Opening: " .. tostring( Door ) .. " #" .. Door.ID );
			Door:GetController():SetState( Controller.PRIMARY_ACTION, true );
		end
	end

	-- Announce rescue
	self:UpdateMessage( "A rescue craft has arrived! Depart as soon as possible!" );

end

function ZombieDefense1:CreateDefender()

	print( "Adding defender..." );
	local Defender = CreateAHuman( "Gordon" );
	Defender.Team = 0;
	Defender.AIMode = Actor.AIMODE_NONE
	self:SetPlayerBrain( Defender, 0 );
	self:EquipDefender( Defender )
	return Defender

end

function ZombieDefense1:DropDefender()

	-- Announce reinforcements
	if self.Wave then
		self:UpdateMessage( "A reinforcement has arrived!" );
	end

	local Crate = CreateACRocket( "Drop Crate", "Dummy.rte" );
	Crate.Team = 0;
	Crate.Pos = Vector( self.ReinforceArea:GetRandomPoint().X, -50 );
	local Defender = self:CreateDefender();
	Crate:AddInventoryItem( Defender );
	MovableMan:AddActor( Crate );
	return Defender;

end

function ZombieDefense1:EquipDefender( Defender )

	for _, Weapon in ipairs( self.Weapons ) do
		if not Defender:HasObject( Weapon ) then
			Defender:AddInventoryItem( CreateHDFirearm( Weapon ) );
		end
	end

end

function ZombieDefense1:CheckRewards()

	local Kills = self.ZombieKilled

	-- Find the highest earned level
	local Level, New, Inv = -1, nil, nil
	for KillReq, Weapon in pairs( self.WeaponRewards ) do
		if KillReq > self.WeaponLevel and Kills > KillReq then
			Level = KillReq;
			New = Weapon.New;
			Inv = Weapon.Inv;
		end
	end

	if New and Inv then

		-- If a new weapon level was reached, set it up

		-- Remember this level so we don't reach the same level again
		self.WeaponLevel = Level;
		-- Replace the weapon list with the short list for this level
		self.Weapons = Inv;

		-- Replace the inventories
		for Defender in MovableMan.Actors do
			for _, Weapon in pairs( self.Weapons ) do
				if Defender.Team == 0 and Defender.ClassName == "AHuman" then
					self:EquipDefender( Defender );
				end
			end
		end

		-- Announce the new weapon
		print( "Earned the " .. New .. "!" );
		self:UpdateMessage( "You've earned the " .. New );

	end

end

function ZombieDefense1:SpawnZombie()

	-- Count the zombies
	local ZombieCount = 0;
	for Zombie in MovableMan.Actors do
		if Zombie.Team == 1 and Zombie.ClassName == "AHuman" then
			ZombieCount = ZombieCount + 1;
		end
	end

	-- Determine how many have died
	local Kills = ( self.ZombieCount - ZombieCount )
	if Kills > 0 then
		self.ZombieKilled = self.ZombieKilled + Kills;
		self.WaveKilled = self.WaveKilled + Kills;
		self:CheckRewards()
		if self.WaveKilled > self.WaveCap - 3 then
			self:StopWave();
		end
	end

	-- Spawn a zombie
	if ZombieCount < self.ZombieCap and self.WaveSpawned < self.WaveCap then
		local ZombieType = self.ZombieTypes[ math.random( #self.ZombieTypes ) ];
		--print( "Spawning " .. ZombieType );
		local Zombie = CreateAHuman( ZombieType );
		Zombie:AddAISceneWaypoint( self.GoalArea:GetRandomPoint() );
		Zombie.AIMode = Actor.AIMODE_GOTO;
		--Zombie.AIMode = Actor.AIMODE_PATROL;
		Zombie.Pos = self:PickZombieStart( Zombie )
		Zombie.Team = 1;
		if #self.ZombieWeapons > 0 then
			Zombie:AddInventoryItem( CreateHDFirearm( self.ZombieWeapons[ math.random( #self.ZombieWeapons ) ] ) );
		end
		if #self.ZombieBombs > 0 then
			Zombie:AddInventoryItem( CreateTDExplosive( self.ZombieBombs[ math.random( #self.ZombieBombs ) ] ) );
		end
		MovableMan:AddActor( Zombie );
		ZombieCount = ZombieCount + 1;
		self.ZombieSpawned = self.ZombieSpawned + 1;
		self.WaveSpawned = self.WaveSpawned + 1;

	end

	-- Update the count and resent the timer
	self.ZombieCount = ZombieCount;
	self.ZombieSpawner:Reset();

end

function ZombieDefense1:PickZombieStart()

	local SpawnArea;
	if self.ZombieRight and ( math.random() < self.ZombieRight ) then
		SpawnArea = self.ZombieSpawnRight;
	else
		SpawnArea = self.ZombieSpawnLeft;
	end

	return SceneMan:MovePointToGround( SpawnArea:GetRandomPoint(), 4, 0 );

end

function pt( v )
	for key, value in pairs( v ) do
		if type( value ) == "string" or type( value ) == "number" then
			print( key .. " = [" .. value .. "]" )
		elseif type( value ) == "table" then
			print( key .. " is table" )
		else
			print( key .. " is " .. type( value ) )
		end
	end
end